from typing import Any, Generic, Iterator, Optional, Type, TypeVar, TYPE_CHECKING

import httpx

if TYPE_CHECKING:
    from .client import Client

T = TypeVar("T")


class SimpleResponse:
    """A simple http response intended for cases where the response body is not needed."""

    raw_request: httpx.Request
    raw_response: httpx.Response

    def __init__(self, raw_response: httpx.Response) -> None:
        self.raw_request = raw_response.request
        self.raw_response = raw_response

    @property
    def call_id(self) -> str:
        return self.raw_response.headers.get("x-aryn-call-id")

    @property
    def status_code(self) -> int:
        return self.raw_response.status_code


class Response(SimpleResponse, Generic[T]):
    raw_request: httpx.Request
    raw_response: httpx.Response
    value: T

    def __init__(self, raw_response: httpx.Response, value: T) -> None:
        super().__init__(raw_response)
        self.value = value


class PaginatedResponse(Generic[T]):
    curr_raw_response: httpx.Response
    curr_page: list[T]
    _next_token: Optional[str]

    def __init__(
        self,
        client: "Client",
        first_response: httpx.Response,
        response_type: Type[T],
        list_key: str,
        request_args: list[Any],
        request_kwargs: dict[str, Any],
    ) -> None:
        self._response_type = response_type
        self._client = client
        self._list_key = list_key
        self._request_args = request_args
        self._request_kwargs = request_kwargs

        self._process_response(first_response)

    def _process_response(self, response: httpx.Response) -> None:
        self.curr_raw_response = response
        json_res = response.json()
        self._next_token = json_res.get("next_page_token")
        list_of_dicts = json_res.get(self._list_key, [])

        # TODO: Would like to get this to work with TypeAdapter[list[T]], but
        # that doesn't seem to work.
        self.curr_page = [self._response_type(**v) for v in list_of_dicts]

    def _get_next_page(self) -> None:
        if self._next_token is None:
            self.curr_page = []
            return

        curr_json = self._request_kwargs.get("params", {}) or {}
        curr_json.update({"page_token": self._next_token})
        self._request_kwargs["params"] = curr_json

        new_request = self._client.client.build_request(*self._request_args, **self._request_kwargs)

        self.curr_raw_response = self._client._make_raw_request(new_request)
        self._process_response(self.curr_raw_response)

    @property
    def call_id(self) -> str:
        return self.curr_raw_response.headers.get("x-aryn-call-id")

    @property
    def status_code(self) -> int:
        return self.curr_raw_response.status_code

    def get_all(self) -> list[T]:
        return list(self)

    def iter_page(self) -> Iterator[Response[list[T]]]:
        yield Response[list[T]](self.curr_raw_response, self.curr_page)
        while self._next_token is not None:
            self._get_next_page()
            yield Response[list[T]](self.curr_raw_response, self.curr_page)

    def __iter__(self) -> Iterator[T]:
        yield from self.curr_page
        while self._next_token is not None:
            self._get_next_page()
            yield from self.curr_page



from typing import Annotated, Any, Literal, Optional, Union
from pydantic import AliasChoices, BaseModel, Field, JsonValue, PlainValidator, WrapSerializer


class DocumentMetadata(BaseModel):
    """ """

    account_id: str = Field(description="The account id containing the Document.")
    doc_id: str = Field(description="The unique id for the Document.")
    docset_id: str = Field(description="The unique id for the DocSet containing the Document.")
    created_at: str = Field(description="The time at which this document was added to the system.")
    name: str = Field(description="The name of this Document.")
    size: int = Field(description="The size of the Document in bytes.")
    content_type: str = Field(description="The content type of the source Document.")
    # TODO: What do we use properties for here?
    properties: Optional[dict] = Field(default=None, description="Additional properties for the Document.")


class Element(BaseModel):
    """ """

    id: str = Field(description="The unique id for the Element.")
    type: str = Field(description="The type of the Element.")
    text_representation: Optional[str] = Field(description="The text representation of the Element.")
    embedding: Optional[list[float]] = Field(description="The vector embedding of the Element.")
    properties: dict[str, Any] = Field(default={}, description="A map of properties for the Element.")
    bbox: tuple[float, float, float, float] = Field(description="The bounding box of the Element.")


class Document(BaseModel):
    """ """

    id: str = Field(description="The unique id for the Document.", validation_alias=AliasChoices("id", "doc_id"))
    elements: list[Element] = Field(default=[], description="The elements contained in the Document.")
    properties: dict[str, Any] = Field(default={}, description="A map of properties for the Document.")
    binary_data: Optional[str] = Field(
        default=None, description="The binary content of the document, encoded as a base64 string."
    )


class JSONPointer:
    """Represents a JSON Pointer as defined in RFC 6901.

    This is a standard way to reference a specific value in a JSON document.

    A JSON Pointer is a / delimited string, where each component is either an
    object key or an array index. When adding an element, you can also use '-'
    to indicate the end of an array. '~' and '/' must be escaped: ~ is encoded
    as ~0 and / is encoded as ~1.

    The easiest way to construct a JSONPointer instance is via JSONPointer.parse.

    Examples:
      - JSONPointer.parse("/foo/bar") -> JSONPointer(["foo", "bar"])
      - JSONPointer.parse("/foo/bar/3") -> JSONPointer(["foo", "bar", "3"])
    """

    parts: list[str]

    def __init__(self, parts: list[str]):
        self.parts = parts

    def __str__(self):
        return "/" + "/".join(p.replace("~", "~0").replace("/", "~1") for p in self.parts)

    @classmethod
    def parse(cls, value: Union[str, list[str], "JSONPointer"]) -> "JSONPointer":
        """Creates a JSONPointer from a string.

        Also handles a list o strings (like the constructor), and a
        JSONPointer instance, which is returned as-is. This is for Pydantic
        compatibility.
        """
        if isinstance(value, JSONPointer):
            return value
        elif isinstance(value, list):
            return cls(value)

        assert isinstance(value, str)

        if value == "":
            return cls([])

        if not value.startswith("/"):
            raise ValueError("JSON Pointer must start with a '/'")

        value = value[1:]
        parts = [] if len(value) == 0 else value.split("/")

        try:
            dash_idx = parts.index("-")
        except ValueError:
            dash_idx = -1

        if dash_idx != -1 and dash_idx != len(parts) - 1:
            raise ValueError("Array end marker '-' must be at the end of the path.")

        return cls([p.replace("~1", "/").replace("~0", "~") for p in parts])


# This is a type that can be used in a Pydantic model that will be used to convert to/from
# JSONPointer strings.
JSONPointerType = Annotated[
    JSONPointer,
    PlainValidator(JSONPointer.parse),
    WrapSerializer(lambda v, handler: str(v) if isinstance(v, JSONPointer) else handler(v)),
]


class JSONPatchOperationBase(BaseModel):
    """Base class to represent a JSON Patch operation.

    JSONPatch is a mechanism for representing updates to a JSON document
    defined at https://jsonpatch.com.
    """

    pass


class AddOperation(JSONPatchOperationBase):
    op: Literal["add"] = "add"
    path: JSONPointerType
    value: JsonValue


class RemoveOperation(JSONPatchOperationBase):
    op: Literal["remove"] = "remove"
    path: JSONPointerType


class ReplaceOperation(JSONPatchOperationBase):
    op: Literal["replace"] = Field(
        default="replace", description="The type of JSON Patch operation. Currently only 'replace' is supported"
    )
    path: JSONPointerType = Field(
        description="a [JSON Pointer](https://datatracker.ietf.org/doc/html/rfc6901) object identifying the value to be replaced."
    )
    value: JsonValue = Field(description="The new value to replace the existing value.")


class FieldUpdates(BaseModel):
    operations: list[ReplaceOperation] = Field(
        description="A list of updates to a document that will be applied sequentially."
    )



from typing import Annotated, Literal, Optional
from pydantic import BaseModel, Field, model_validator


class SearchRequest(BaseModel):
    query: Annotated[Optional[str], Field(description="The query to search for")] = None
    query_type: Annotated[
        Literal["keyword", "lexical", "vector", "hybrid"], Field(description="The type of query to perform.")
    ] = "lexical"
    properties_filter: Annotated[
        Optional[str], Field(description="A filter to apply to the properties of the documents.")
    ] = None
    return_type: Annotated[Literal["doc", "element"], Field(description="The type of result to return.")] = "doc"
    include_fields: Annotated[
        Optional[list[str]],
        Field(
            description="When specified, the server performs a projection to return only the fields specified by this list of JsonPaths (see RFC 9535) given as a list of strings.",
        ),
    ] = None
    debug_info: Annotated[
        Optional[bool],
        Field(
            description="When `True`, the server returns additional debug information (the embedding calculated for a vector or hybrid query) in the response.",
        ),
    ] = False
    include_element_embedding: Annotated[
        Optional[bool],
        Field(
            description='When `False`, and return_type is "element", the server does not return the embeddings of elements in the response.',
        ),
    ] = False

    @model_validator(mode="after")
    def check_query_or_properties_filter_present(self):
        if self.query is None and self.properties_filter is None:
            raise ValueError("Either query or properties_filter must be provided")
        return self

    @model_validator(mode="after")
    def require_query_when_query_type_uses_vector_search(self):
        if self.query is None and self.query_type in ("vector", "hybrid"):
            raise ValueError("query must be provided when query_type is vector or hybrid")
        return self

    @model_validator(mode="after")
    def disallow_empty_string_when_query_set(self):
        if self.query == "":
            raise ValueError("query may be omitted, but must not be an empty string")
        return self

    class Config:
        # Ensure that extra fields are not allowed
        extra = "forbid"


class SearchResponse(BaseModel):
    results: Annotated[
        list,
        Field(
            description="The list of results returned by the query. These are Json objects representing the documents or elements that matched the query, possibly filtered by the optional projection specified by include_fields."
        ),
    ]

    # the field query_embedding below is to allow users to debug their queries and/or their docset, and is only present
    # when query_type is vector or hybrid and debug_info is True
    query_embedding: Annotated[
        Optional[list[float]],
        Field(
            description="The embedding calculated for the query, if the query type is vector or hybrid and `debug_info` is True.",
        ),
    ] = None

    next_page_token: Annotated[
        Optional[str],
        Field(
            description="Reserved for future use. Will be a pagination token that can be used to retrieve the next page of results.",
        ),
    ] = None



from contextlib import nullcontext
import json
import logging
import mimetypes
from os import PathLike
from typing import Any, BinaryIO, cast, ContextManager, Iterator, Optional, Tuple, Type, TypeVar, Union

import httpx
from httpx import Request
from httpx_sse import connect_sse
from pydantic import JsonValue

from .config import ArynConfig
from .exceptions import ArynSDKException
from .response import Response, PaginatedResponse, SimpleResponse
from .tasks import AsyncTask
from ..types.docset import DocSetMetadata, DocSetUpdate
from ..types.document import Document, DocumentMetadata, FieldUpdates
from ..types.prompt import PromptType
from ..types.query import LogicalPlan, Query, QueryResult, QueryTraceDoc, QueryEvent, QueryEventType
from ..types.schema import Schema
from ..types.search import SearchRequest, SearchResponse
from ..types.task import AsyncTaskMap
from ..types.transforms import TransformResponse

ResponseType = TypeVar("ResponseType")


class Client:

    def __init__(
        self,
        aryn_url: str = "https://api.aryn.ai",
        aryn_api_key: Optional[str] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: float = 120.0,
    ) -> None:
        self.aryn_url = aryn_url

        self.config = ArynConfig(aryn_api_key=aryn_api_key, aryn_url=aryn_url)

        headers = (extra_headers or {}) | {"Authorization": f"Bearer {self.config.api_key()}"}
        self.client = httpx.Client(base_url=self.config.aryn_url(), headers=headers, timeout=timeout)

    def _make_raw_request(self, req: Request) -> httpx.Response:
        res = self.client.send(req)
        if res.status_code >= 300:
            raise ArynSDKException(res)

        return res

    def _make_request(
        self,
        req: Request,
        response_type: Type[ResponseType],
    ) -> Response[ResponseType]:
        res = self._make_raw_request(req)

        return Response(raw_response=res, value=response_type(**res.json()))

    def _make_paginated_request(
        self, req: Request, responseType: Type[ResponseType], list_key: str, *request_args, **request_kwargs
    ) -> PaginatedResponse[ResponseType]:
        res = self._make_raw_request(req)
        return PaginatedResponse(
            client=self,
            first_response=res,
            response_type=responseType,
            list_key=list_key,
            request_args=list(request_args),
            request_kwargs=request_kwargs,
        )

    # ----------------------------------------------
    # DocSet APIs
    # ----------------------------------------------

    def create_docset(
        self,
        *,
        name: str,
        properties: Optional[dict[str, JsonValue]] = None,
        schema: Optional[Schema] = None,
        prompts: Optional[dict[PromptType, str]] = None,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> Response[DocSetMetadata]:
        json_body = {"name": name, "properties": properties, "prompts": prompts}
        if schema is not None:
            json_body["schema"] = schema.model_dump()

        req = self.client.build_request("POST", "/v1/storage/docsets", json=json_body, headers=extra_headers)
        return self._make_request(req, DocSetMetadata)

    def clone_docset(
        self, *, docset_id: str, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[DocSetMetadata]:
        return self._make_request(
            self.client.build_request("POST", f"/v1/storage/docsets/{docset_id}/clone", headers=extra_headers),
            DocSetMetadata,
        )

    def list_docsets(
        self,
        *,
        name_eq: Optional[str] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> PaginatedResponse[DocSetMetadata]:

        params: dict[str, Any] = {}
        if name_eq is not None:
            params["name_eq"] = name_eq
        if page_size is not None:
            params["page_size"] = page_size
        if page_token is not None:
            params["page_token"] = page_token

        args = ("GET", "/v1/storage/docsets")
        kwargs: dict[str, Any] = {"params": params or None, "headers": extra_headers}
        req = self.client.build_request(*args, **kwargs)
        return self._make_paginated_request(req, DocSetMetadata, "items", *args, **kwargs)

    def get_docset(self, *, docset_id: str, extra_headers: Optional[dict[str, str]] = None) -> Response[DocSetMetadata]:
        return self._make_request(
            self.client.build_request("GET", f"/v1/storage/docsets/{docset_id}", headers=extra_headers), DocSetMetadata
        )

    def update_docset(
        self, *, docset_id: str, update: DocSetUpdate, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[DocSetMetadata]:
        return self._make_request(
            self.client.build_request(
                "PATCH", f"/v1/storage/docsets/{docset_id}", json=update.model_dump(), headers=extra_headers
            ),
            DocSetMetadata,
        )

    def set_readonly_docset(
        self, *, docset_id: str, readonly: bool, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[DocSetMetadata]:
        return self._make_request(
            self.client.build_request(
                "POST", f"/v1/storage/docsets/{docset_id}/readonly/{int(readonly)}", headers=extra_headers
            ),
            DocSetMetadata,
        )

    def delete_docset(
        self, *, docset_id: str, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[DocSetMetadata]:
        return self._make_request(
            self.client.build_request("DELETE", f"/v1/storage/docsets/{docset_id}", headers=extra_headers),
            DocSetMetadata,
        )

    # ----------------------------------------------
    # Document APIs
    # ----------------------------------------------

    def _resolve_add_doc_file(
        self, file: Union[BinaryIO, str, PathLike]
    ) -> Union[BinaryIO, Tuple[str, BinaryIO, Optional[str]]]:
        if isinstance(file, (str, PathLike)):
            if str(file).startswith("s3://"):
                try:
                    import boto3
                except ImportError:
                    raise ImportError("Please install the boto3 library to read from S3 URLs.")

                s3 = boto3.client("s3")
                bucket, key = str(file)[5:].split("/", 1)
                response = s3.get_object(Bucket=bucket, Key=key)
                stream = cast(BinaryIO, response["Body"])
            else:
                stream = open(file, "rb")

            mime_type, _ = mimetypes.guess_type(file)
            file_request: Union[BinaryIO, Tuple[str, BinaryIO, Optional[str]]] = (
                str(file),
                stream,
                mime_type or "application/octet-stream",
            )
        else:
            file_request = file

        # boto3 types don't play well with mypy.
        return file_request  # type: ignore

    # TODO: Better typing of DocParse options.
    def add_doc(
        self,
        *,
        file: Union[BinaryIO, str, PathLike],
        docset_id: str,
        options: Optional[dict[str, Any]] = None,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> Response[DocumentMetadata]:
        file_request: Any

        file_request = self._resolve_add_doc_file(file)
        files: dict[str, Any] = {"file": file_request}

        if options is not None:
            files["options"] = json.dumps(options).encode("utf-8")

        req = self.client.build_request(
            "POST", f"/v1/storage/docsets/{docset_id}/docs", files=files, headers=extra_headers
        )
        return self._make_request(req, DocumentMetadata)

    def _add_doc_async_internal(
        self,
        *,
        file: Union[BinaryIO, str, PathLike],
        docset_id: str,
        options: Optional[dict[str, Any]] = None,
        extra_headers,
    ) -> httpx.Response:

        file_request = self._resolve_add_doc_file(file)
        files: dict[str, Any] = {"file": file_request}
        data: dict[str, Any] = {"options": None}

        if options is not None:
            data["options"] = json.dumps(options).encode("utf-8")

        req = self.client.build_request(
            "POST", f"/v1/async/submit/storage/docsets/{docset_id}/docs", json=data, files=files, headers=extra_headers
        )
        return self._make_raw_request(req)

    def add_doc_async(
        self,
        *,
        file: Union[BinaryIO, str, PathLike],
        docset_id: str,
        options: Optional[dict[str, Any]] = None,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> AsyncTask[DocumentMetadata]:
        res = self._add_doc_async_internal(file=file, docset_id=docset_id, options=options, extra_headers=extra_headers)

        task_id = res.json()["task_id"]
        return AsyncTask(
            client=self,
            task_id=task_id,
            method="POST",
            path=f"/storage/docsets/{docset_id}/docs",
            response_type=DocumentMetadata,
        )

    # TODO: Decide what filtering we want to support here
    def list_docs(
        self,
        *,
        docset_id: str,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> PaginatedResponse[DocumentMetadata]:

        page_param: dict[str, Any] = {}
        if page_size is not None:
            page_param["page_size"] = page_size
        if page_token is not None:
            page_param["page_token"] = page_token

        args = ("GET", f"/v1/storage/docsets/{docset_id}/docs")
        kwargs: dict[str, Any] = {"params": page_param or None, "headers": extra_headers}
        req = self.client.build_request(*args, **kwargs)
        return self._make_paginated_request(req, DocumentMetadata, "items", *args, **kwargs)

    def get_doc(
        self,
        *,
        docset_id,
        doc_id,
        include_elements: bool = True,
        include_binary: bool = False,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> Response[Document]:
        data = {"include_elements": include_elements, "include_binary": include_binary}
        req = self.client.build_request(
            "GET", f"/v1/storage/docsets/{docset_id}/docs/{doc_id}", data=data, headers=extra_headers
        )
        return self._make_request(req, Document)

    def get_doc_binary(
        self,
        *,
        docset_id: str,
        doc_id: str,
        file: Union[BinaryIO, PathLike, str],
        extra_headers: Optional[dict[str, str]] = None,
    ) -> None:
        # TODO: This should really be a streaming response, and I'm not sure
        # that writing it to a file is the best way to handle it, but it simplifies
        # the typing of the response for now.
        req = self.client.build_request(
            "GET", f"/v1/storage/docsets/{docset_id}/docs/{doc_id}/binary", headers=extra_headers
        )
        res = self._make_raw_request(req)

        if isinstance(file, (str, PathLike)):
            cm: ContextManager[BinaryIO] = open(file, "wb")
        else:
            cm = nullcontext(file)

        with cm as file_obj:
            for data in res.iter_bytes():
                file_obj.write(data)

    def update_doc_properties(
        self, *, docset_id: str, doc_id: str, updates: FieldUpdates, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[Document]:
        req = self.client.build_request(
            "PATCH",
            f"/v1/storage/docsets/{docset_id}/docs/{doc_id}/properties",
            json=updates.model_dump(),
            headers=extra_headers,
        )
        return self._make_request(req, Document)

    def delete_doc(
        self, *, docset_id: str, doc_id: str, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[DocumentMetadata]:
        return self._make_request(
            self.client.build_request(
                "DELETE", f"/v1/storage/docsets/{docset_id}/docs/{doc_id}", headers=extra_headers
            ),
            DocumentMetadata,
        )

    # ----------------------------------------------
    # Search APIs
    # ----------------------------------------------
    def search(
        self,
        *,
        docset_id: str,
        query: SearchRequest,
        page_size: Optional[int] = None,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> Response[SearchResponse]:
        params = {"page_size": page_size} if page_size is not None else None
        req = self.client.build_request(
            "POST", f"/v1/query/search/{docset_id}", params=params, json=query.model_dump(), headers=extra_headers
        )
        return self._make_request(req, SearchResponse)

    # ----------------------------------------------
    # Query APIs
    # ----------------------------------------------

    def generate_plan(self, *, query: Query, extra_headers: Optional[dict[str, str]] = None) -> Response[LogicalPlan]:
        req = self.client.build_request(
            "POST",
            "/v1/query/plan",
            json=query.model_dump(),
            headers=extra_headers,
        )
        return self._make_request(req, LogicalPlan)

    def edit_plan(
        self, *, query: Query, feedback: str, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[LogicalPlan]:
        body = {"query": query.model_dump(), "feedback": feedback}

        req = self.client.build_request(
            "PATCH",
            "/v1/query/plan",
            json=body,
            headers=extra_headers,
        )
        return self._make_request(req, LogicalPlan)

    def _query_streamed(self, *, method: str, url: str, kwargs: dict[str, Optional[dict[str, str]]]):
        with connect_sse(self.client, method, url, **kwargs) as event_source:
            for sse in event_source.iter_sse():
                value: Any

                if sse.event == QueryEventType.PLAN:
                    value = LogicalPlan.model_validate_json(sse.data)
                elif sse.event == QueryEventType.RESULT_DOC:
                    value = Document.model_validate_json(sse.data)
                elif sse.event == QueryEventType.TRACE_DOC:
                    value = QueryTraceDoc.model_validate_json(sse.data)
                else:
                    # TODO: result events can include numeric data, but there
                    # is no type marker to indicate the type. For now we will
                    # just return as a string.
                    value = sse.data

                yield QueryEvent(event_type=QueryEventType(sse.event), data=value)

    def query(
        self, *, query: Query, extra_headers: Optional[dict[str, str]] = None
    ) -> Union[Response[QueryResult], Iterator[QueryEvent]]:

        method = "POST"
        url = "/v1/query"
        kwargs = {
            "json": query.model_dump(),
        }
        if extra_headers:
            kwargs["headers"] = extra_headers

        if not query.stream:
            req = self.client.build_request(method, url, **kwargs)
            return self._make_request(req, QueryResult)
        else:
            return self._query_streamed(method=method, url=url, kwargs=kwargs)

    # ----------------------------------------------
    # Transform APIs
    # ----------------------------------------------

    def extract_properties(
        self, *, docset_id: str, schema: Schema, extra_headers: Optional[dict[str, str]] = None
    ) -> Response[TransformResponse]:
        req = self.client.build_request(
            "POST",
            "/v1/jobs/extract-properties",
            params={"docset_id": docset_id},
            json=schema.model_dump(),
            headers=extra_headers,
        )
        return self._make_request(req, TransformResponse)

    def extract_properties_async(
        self, *, docset_id: str, schema: Schema, extra_headers: Optional[dict[str, str]] = None
    ) -> AsyncTask[TransformResponse]:
        req = self.client.build_request(
            "POST",
            "/v1/async/submit/jobs/extract-properties",
            params={"docset_id": docset_id},
            json=schema.model_dump(),
            headers=extra_headers,
        )

        res = self._make_raw_request(req)
        task_id = res.json()["task_id"]
        return AsyncTask(
            client=self,
            task_id=task_id,
            method="POST",
            path="/jobs/extract-properties",
            response_type=TransformResponse,
        )

    def delete_properties(
        self,
        *,
        docset_id: str,
        property_names: list[str],
        extra_headers: Optional[dict[str, str]] = None,
    ) -> Response[TransformResponse]:

        req = self.client.build_request(
            "POST",
            "/v1/jobs/delete-properties",
            params={"docset_id": docset_id},
            json={"names": property_names},
            headers=extra_headers,
        )
        return self._make_request(req, TransformResponse)

    def delete_properties_async(
        self,
        *,
        docset_id: str,
        property_names: list[str],
        extra_headers: Optional[dict[str, str]] = None,
    ) -> AsyncTask[TransformResponse]:

        req = self.client.build_request(
            "POST",
            "/v1/async/submit/jobs/delete-properties",
            params={"docset_id": docset_id},
            json={"names": property_names},
            headers=extra_headers,
        )

        res = self._make_raw_request(req)
        task_id = res.json()["task_id"]
        return AsyncTask(
            client=self,
            task_id=task_id,
            method="POST",
            path="/jobs/delete-properties",
            response_type=TransformResponse,
        )

    # ----------------------------------------------
    # Async task APIs
    # ----------------------------------------------

    def _get_task_and_filters(self, task: Union[AsyncTask, str]) -> Tuple[str, Optional[str], Optional[str]]:
        method_filter = None
        path_filter = None

        if isinstance(task, AsyncTask):
            task_id = task.task_id

            if (method := task.method) is not None:
                method_filter = method

            if (path := task.path) is not None:
                path_filter = path
        else:
            task_id = task

        return task_id, method_filter, path_filter

    def list_async_tasks(self, extra_headers: Optional[dict[str, str]] = None) -> Response[AsyncTaskMap]:
        req = self.client.build_request("GET", "/v1/async/list", headers=extra_headers)
        return self._make_request(req, AsyncTaskMap)

    def cancel_async_task(
        self, task: Union[AsyncTask, str], extra_headers: Optional[dict[str, str]] = None
    ) -> SimpleResponse:
        task_id, method_filter, path_filter = self._get_task_and_filters(task)

        req = self.client.build_request(
            "POST",
            f"/v1/async/cancel/{task_id}",
            params={"method_filter": method_filter, "path_filter": path_filter},
            headers=extra_headers,
        )

        res = self._make_raw_request(req)
        return SimpleResponse(res)

    def _get_async_result_internal(
        self, task: Union[AsyncTask, str], extra_headers: Optional[dict[str, str]] = None
    ) -> httpx.Response:
        task_id, method_filter, path_filter = self._get_task_and_filters(task)

        req = self.client.build_request(
            "GET",
            f"/v1/async/result/{task_id}",
            params={"method_filter": method_filter, "path_filter": path_filter},
            headers=extra_headers,
        )

        return self._make_raw_request(req)

    def get_async_result(
        self, task: Union[AsyncTask, str], extra_headers: Optional[dict[str, str]] = None
    ) -> Union[SimpleResponse, Response]:
        res = self._get_async_result_internal(task, extra_headers=extra_headers)

        if res.status_code == 200:
            if res.headers.get("Content-Type").lower() == "application/json":
                content = res.json()
            else:
                content = res.content

            return Response(res, content)

        elif res.status_code == 202:
            return SimpleResponse(res)

        # This should be unreachable, as other status codes should be handled by _make_raw_request
        logging.error(f"Unexpected status code {res.status_code} for async task {task}")
        raise ArynSDKException(res)



from ..client.partition import (
    partition_file,
    partition_file_async_submit,
    partition_file_async_result,
    partition_file_async_cancel,
    partition_file_async_list,
    tables_to_pandas,
    tables_to_html,
    table_elem_to_dataframe,
    table_elem_to_html,
    convert_image_element,
    BoolFlag,
    PartitionError,
    PartitionTaskError,
    PartitionTaskNotFoundError,
)
from ..client.art import draw_with_boxes

__all__ = [
    "partition_file",
    "table_elem_to_dataframe",
    "table_elem_to_html",
    "tables_to_pandas",
    "tables_to_html",
    "draw_with_boxes",
    "convert_image_element",
    "BoolFlag",
    "PartitionError",
    "PartitionTaskError",
    "PartitionTaskNotFoundError",
    "partition_file_async_submit",
    "partition_file_async_result",
    "partition_file_async_cancel",
    "partition_file_async_list",
]



from typing import Any, Optional
from pydantic import AliasChoices, BaseModel, Field

# TODO: Some kind of TypeAlias to get this to work with Sycamore schemas as well. Uggh.


class SchemaProperty(BaseModel):
    """Represents a field in a DocSet schema."""

    name: str = Field(description="The name of the property.")
    property_type: str = Field(
        description="The type of the field.", validation_alias=AliasChoices("field_type", "property_type")
    )

    @property
    def field_type(self) -> str:
        """Alias for property_type."""
        return self.property_type

    default: Optional[Any] = Field(default=None, description="The default value for the property.")
    description: Optional[str] = Field(default=None, description="A natural language description of the property.")
    examples: Optional[list[Any]] = Field(default=None, description="A list of example values for the property.")


SchemaField = SchemaProperty


class Schema(BaseModel):
    """Represents the schema of a DocSet."""

    properties: list[SchemaProperty] = Field(
        description="A list of properties belong to this schema.",
        validation_alias=AliasChoices("properties", "fields"),
    )

    @property
    def fields(self) -> list[SchemaProperty]:
        """Alias for properties."""
        return self.properties


class SchemaPropertyNames(BaseModel):
    """Represents the names of properties belonging to a DocSet schema."""

    names: list[str] = Field(description="A list of names of properties that belong to a schema.")



"""
Functionality to add multiple files/directories to DocStore

The easy way to call this is:

r = add_docs(docset_id, pathlist)

However, it's likely that a filter is needed, as so:

f = Filter().include_glob("*.pdf")
r = add_docs(docset_id, pathlist, f)

The underlying classes can be used, too:

pdf = Filter().include_regex("[.]pdf$")
doc = Filter().include_regex("[.]docx$")
a = DocAdder(docset_id)
a.add(pathlist1, pdf).add(pathlist2, doc)
r = a.finish()
"""

import os
import re
import stat
import time
import json
import logging
import threading
from queue import Empty, Queue
from typing import Any, Optional, Union
from contextlib import suppress

from .client import Client
from .exceptions import ArynSDKException

import httpx


g_logger = logging.getLogger("DocAdder")


class Filter:
    """Uses an include pattern list and an exclude pattern list to decide
    if a given path should be allowed."""

    def __init__(self) -> None:
        self.includes = PatList()
        self.excludes = PatList()

    def include_regex(self, r: Union[str, re.Pattern]) -> "Filter":
        self.includes.add(r)
        return self

    def exclude_regex(self, r: Union[str, re.Pattern]) -> "Filter":
        self.excludes.add(r)
        return self

    def include_glob(self, g: str) -> "Filter":
        self.includes.add(glob2re(g))
        return self

    def exclude_glob(self, g: str) -> "Filter":
        self.excludes.add(glob2re(g))
        return self

    def check(self, path) -> bool:
        if self.includes and not self.includes.any_search(path):
            return False
        if self.excludes and self.excludes.any_search(path):
            return False
        return True


class DocAdder:
    """Client for DocStore to add files/directories to a DocSet.  Once
    instantiated, call add() one or more times, then call finish().  It
    may take a long time as it waits for all documents to finish."""

    def __init__(self, client: Client, docset_id: str) -> None:
        self.client = client
        self.docset_id = docset_id
        self.maxtasks = 1000
        self.tasks: dict[str, str] = {}
        self.done: dict[str, dict[str, str]] = {}

        userdata = {"TargetDocSet": docset_id}

        self.extra_headers = {
            "User-Agent": "AddDocs/0.1",
            "X-Aryn-Userdata": json.dumps(userdata),
        }

        self.submit = f"/v1/async/submit/docstore/docsets/{self.docset_id}/docs"
        self.options: dict[str, Any] = {
            "use_ocr": True,
            "extract_table_structure": True,
        }

        self.scan_q: Queue = Queue(100)
        self.doc_q: Queue = Queue(100)
        self.scan_thr = threading.Thread(target=self._dirscan_worker, name="DirScanWorker")
        self.doc_thr = threading.Thread(target=self._docstore_worker, name="AddDocsWorker")
        self.scan_thr.start()
        self.doc_thr.start()

    def add(
        self, paths: list[str], filter: Optional[Filter] = None, options: Optional[dict[str, Any]] = None
    ) -> "DocAdder":
        """Add multiple files/directories to the DocSet."""
        if filter is None:
            filter = Filter()
        if options is None:
            options = self.options
        self.scan_q.put(ScanRec(paths, filter, options))
        return self

    def finish(self) -> dict[str, dict[str, str]]:
        """Wait for all documents to finish and return map of results."""
        self.scan_q.put(None)
        self.scan_thr.join()
        self.doc_q.put(None)
        self.doc_thr.join()
        return self.done

    def empty(self) -> bool:
        """Indicate if the internal queues and state are empty."""
        return self.scan_q.empty() and self.doc_q.empty() and not self.tasks

    def progress(self) -> tuple[int, int]:
        """Return number of tasks pending and done as a tuple."""
        return len(self.tasks), len(self.done)

    def cancel_all(self) -> None:
        """This is mostly for testing."""
        for task_id in self.tasks.keys():
            self._cancel_one(task_id)

    def _dirscan_worker(self) -> None:
        while True:
            rec = self.scan_q.get()
            if rec is None:
                break
            for path in rec.paths:
                st = os.stat(path)
                if stat.S_ISDIR(st.st_mode):
                    self._dir_recurse(path, rec.filter, rec.options, 0)
                elif stat.S_ISREG(st.st_mode):
                    if rec.filter.check(path):
                        self.doc_q.put(DocRec(path, rec.options))

    def _dir_recurse(self, dir: str, filt: Filter, opts: dict[str, Any], level: int) -> None:
        if level > 80:  # kernel limits to 40 symlink resolutions per path
            raise RecursionError()
        with os.scandir(dir) as scan:
            ents = list(scan)
        ents.sort(key=lambda e: e.name)
        for ent in ents:
            if ent.is_dir():
                self._dir_recurse(ent.path, filt, opts, level + 1)
            elif ent.is_file():
                if filt.check(ent.path):
                    self.doc_q.put(DocRec(ent.path, opts))

    def _docstore_worker(self) -> None:
        qok = True
        while self.tasks or qok:
            idle = True
            while qok and (len(self.tasks) < self.maxtasks):
                try:
                    rec = self.doc_q.get_nowait()
                    if rec is None:
                        qok = False
                    else:
                        self._submit_one(rec.path, rec.options)
                        idle = False
                except Empty:
                    break
            if idle:
                time.sleep(1.0)
            self._poll_tasks()

    def _poll_tasks(self) -> None:
        nap = 5.0 / max(1, len(self.tasks))
        nap = max(0.001, min(1.0, nap))
        drops: list[str] = []
        for task_id, path in self.tasks.items():
            try:
                r = self.client._get_async_result_internal(task=task_id, extra_headers=self.extra_headers)
                if r.status_code == 202:  # still pending
                    time.sleep(nap)
                elif r.status_code == 200:  # OK
                    data = r.json()
                    doc_id = data["doc_id"]
                    g_logger.info(f"Finished {task_id} {path} {doc_id}")
                    self.done[path] = {"status": "ok", "doc_id": doc_id}
                    drops.append(task_id)
                else:
                    msg = msg_from_resp(r)
                    g_logger.info(f"Got {r.status_code} for {task_id} {msg}")
                    self.done[path] = {"status": "error", "error_response": msg}
                    drops.append(task_id)
            except (ArynSDKException, httpx.RequestError) as e:
                g_logger.warning(f"Exception for {task_id} {e}")
                self.done[path] = {"status": "error", "error_response": str(e)}
                drops.append(task_id)
        for key in drops:  # avoid modifying while iterating
            del self.tasks[key]

    def _submit_one(self, path: str, opts: dict[str, Any]) -> None:
        g_logger.info(f"Submit {path}")
        try:
            r = self.client._add_doc_async_internal(
                file=path, docset_id=self.docset_id, options=opts, extra_headers=self.extra_headers
            )
            if r.status_code == 202:  # pending
                data = r.json()
                task_id = data["task_id"]
                self.tasks[task_id] = path
                g_logger.info(f"Map {task_id} -> {path}")
            else:
                msg = msg_from_resp(r)
                g_logger.warning(f"Got {r.status_code} from {self.submit} {path} {msg}")
                self.done[path] = {"status": "error", "error_response": msg}
        except (ArynSDKException, httpx.RequestError) as e:
            g_logger.warning(f"Exception submitting {path} {e}")

    def _cancel_one(self, task_id) -> None:
        r = self.client.cancel_async_task(task=task_id, extra_headers=self.extra_headers)
        if r.status_code == 200:  # OK
            g_logger.info(f"Cancelled {task_id}")
        elif r.status_code == 404:  # not found
            g_logger.info(f"Cancel too late for {task_id}")
        else:
            g_logger.warning(f"Failed to cancel {task_id}")


def msg_from_resp(resp: httpx.Response) -> str:
    msg = f"generic {resp.status_code}"
    if resp.content.startswith(b'{"'):
        with suppress(KeyError, ValueError):
            data = resp.json()
            msg = data["error"]
    return msg


class PatList:
    """Represents a list of regular expressions to be run with search()."""

    def __init__(self) -> None:
        self.pats: list[re.Pattern] = []

    def __bool__(self) -> bool:
        return bool(self.pats)

    def add(self, r: Union[str, re.Pattern]) -> "PatList":
        if isinstance(r, re.Pattern):
            self.pats.append(r)
        else:
            self.pats.append(re.compile(r))
        return self

    def any_search(self, s: str) -> bool:
        for pat in self.pats:
            if pat.search(s):
                return True
        return False


class ScanRec:
    __slots__ = ("paths", "filter", "options")
    paths: list[str]
    filter: Filter
    options: dict[str, Any]

    def __init__(self, paths: list[str], filter: Filter, options: dict[str, Any]) -> None:
        self.paths = paths
        self.filter = filter
        self.options = options


class DocRec:
    __slots__ = ("path", "options")
    path: str
    options: dict[str, Any]

    def __init__(self, path: str, options: dict[str, Any]) -> None:
        self.path = path
        self.options = options


def add_docs(
    client: Client, docset_id: str, paths: list[str], filter: Optional[Filter] = None
) -> dict[str, dict[str, str]]:
    """
    All-in-one function to add a batch of files and directories to DocStore.
    """
    return DocAdder(client, docset_id).add(paths, filter).finish()


_stars = re.compile(r"[*]+")


def glob2re(glob_str: str) -> re.Pattern:
    """
    From a glob like *.txt make a regex suitable for re.search().
    No current support for **.
    """
    input = _stars.sub("*", glob_str)
    regex = ""
    inclass = False
    n = len(input)
    for idx in range(n):
        ch = input[idx]
        if inclass:
            regex += ch
            if ch == "]":
                inclass = False
        else:
            if ch == "*":
                regex += ".*"
            elif ch == "?":
                regex += "."
            elif ch == "[":
                if input.find("]", idx + 1) >= 0:
                    regex += "["
                    inclass = True
                else:
                    regex += r"\["
            else:
                regex += re.escape(ch)
    regex = f"^{regex}$"
    if regex.startswith("^.*"):
        regex = regex[3:]
    if regex.endswith(".*$"):
        regex = regex[:-3]
    return re.compile(regex)



from os import PathLike
from typing import BinaryIO, Union
from aryn_sdk.client.partition import ARYN_DOCPARSE_URL
import pytest
import json
import time
from pathlib import Path
import logging

from aryn_sdk.partition import (
    partition_file,
    partition_file_async_submit,
    partition_file_async_result,
    partition_file_async_cancel,
    partition_file_async_list,
    PartitionError,
    PartitionTaskNotFoundError,
    BoolFlag,
    convert_image_element,
    tables_to_pandas,
)
from httpx import HTTPStatusError

RESOURCE_DIR = Path(__file__).parent / "resources"
ASYNC_TIMEOUT = 60 * 5  # 5 minutes in seconds


# Unit tests
@pytest.mark.parametrize(
    "pdf, kwargs, response",
    [
        (
            RESOURCE_DIR / "pdfs" / "3m_table.pdf",
            {},
            RESOURCE_DIR / "json" / "3m_output.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "3m_table.pdf",
            {"use_ocr": True},
            RESOURCE_DIR / "json" / "3m_output_ocr.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "3m_table.pdf",
            {"use_ocr": True, "extract_table_structure": True},
            RESOURCE_DIR / "json" / "3m_output_ocr_table.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "3m_table.pdf",
            {"extract_table_structure": True},
            RESOURCE_DIR / "json" / "3m_output_table.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "SPsort.pdf",
            {},
            RESOURCE_DIR / "json" / "SPsort_output.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "SPsort.pdf",
            {"extract_images": True},
            RESOURCE_DIR / "json" / "SPsort_output_images.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "SPsort.pdf",
            {"extract_images": True, "selected_pages": [0]},
            RESOURCE_DIR / "json" / "SPsort_output_images_page0.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "SPsort.pdf",
            {"extract_images": True, "selected_pages": [1]},
            RESOURCE_DIR / "json" / "SPsort_output_images_page1.json",
        ),
    ],
)
def test_partition(pdf, kwargs, response, mocker):

    with open(response, "rb") as f:
        byteses = f.read()

    response_data = json.loads(byteses.decode("utf-8"))
    resp_object = mocker.Mock()
    resp_object.status_code = 200

    # Mock the response from the file,
    resp_object.iter_bytes.return_value = byteses.split(sep=b"\n")

    mocker.patch("httpx.stream").return_value.__enter__.return_value = resp_object

    with open(pdf, "rb") as f:
        if kwargs.get("selected_pages") == [0]:
            with pytest.raises(PartitionError) as einfo:
                new_response = partition_file(f, **kwargs)
            assert "Invalid page number (0)" in str(einfo.value)
        else:
            new_response = partition_file(f, **kwargs)
            assert new_response == response_data


# Integration tests


@pytest.mark.parametrize(
    "pdf, kwargs, response",
    [
        (
            RESOURCE_DIR / "pdfs" / "3m_table.pdf",
            {},
            RESOURCE_DIR / "json" / "3m_output.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "3m_table.pdf",
            {"use_ocr": True, "extract_table_structure": True},
            RESOURCE_DIR / "json" / "3m_output_ocr_table.json",
        ),
        (
            RESOURCE_DIR / "pdfs" / "SPsort.pdf",
            {"extract_images": True, "selected_pages": [1]},
            RESOURCE_DIR / "json" / "SPsort_output_images_page1.json",
        ),
    ],
)
def test_partition_it(pdf, kwargs, response):

    with open(response, "r") as f:
        response_data = json.load(f)

    with open(pdf, "rb") as f:
        new_response = partition_file(f, **kwargs)

    assert len(response_data["elements"]) == len(new_response["elements"])
    for el1, el2 in zip(response_data["elements"], new_response["elements"]):
        assert el1 == el2

    assert response_data["elements"] == new_response["elements"]


def test_partition_with_unsupported_file_format():
    with pytest.raises(PartitionError):
        with open(
            RESOURCE_DIR / "image" / "unsupported-format-test-document-image.heic", "rb"
        ) as f:
            partition_file(f)


def test_partition_it_zero_page():
    with pytest.raises(PartitionError) as einfo:
        with open(RESOURCE_DIR / "pdfs" / "SPsort.pdf", "rb") as f:
            partition_file(f, selected_pages=[0])
    assert "selected_pages must not have empty or zero terms" in str(einfo.value)


def test_partition_it_no_api_key():
    with pytest.raises(HTTPStatusError) as einfo:
        with open(RESOURCE_DIR / "pdfs" / "3m_table.pdf", "rb") as f:
            partition_file(f, aryn_api_key="x")
    assert einfo.value.response.status_code == 403
    assert einfo.value.response.json().get("detail") == "Invalid Aryn API key"


def test_partition_file_auto_rotation():
    expected = json.loads(open(RESOURCE_DIR / "json" / "rotated.json", "r").read())
    actual = partition_file(
        RESOURCE_DIR / "pdfs" / "rotated.pdf",
        output_label_options={"orientation_correction": True},
    )
    assert actual["elements"] == expected["elements"]


def test_partition_file_cancel():
    flag = BoolFlag(True)
    with pytest.raises(Exception):
        partition_file(RESOURCE_DIR / "pdfs" / "SPsort.pdf", cancel_flag=flag)


def test_data_to_pandas():
    with open(RESOURCE_DIR / "json" / "3m_output_ocr_table.json", "r") as f:
        data = json.load(f)
    elts_and_dfs = tables_to_pandas(data)
    assert len(elts_and_dfs) == 5
    df = elts_and_dfs[2][1]
    assert df is not None
    assert df.columns.to_list() == ["(Millions)", "2018", "2017", "2016"]
    assert df["2018"][13] == "134"


def test_convert_img():
    with open(RESOURCE_DIR / "image" / "partitioning_output.json", "r") as f:
        data = json.load(f)
    ims = [e for e in data["elements"] if e["type"] == "Image"]

    jpg_bytes = convert_image_element(ims[0], format="JPEG")
    with open(RESOURCE_DIR / "image" / "jpeg_bytes.jpeg", "rb") as f:
        real_byteses = f.read().strip()
    assert jpg_bytes == real_byteses

    png_str = convert_image_element(ims[0], format="PNG", b64encode=True)
    with open(RESOURCE_DIR / "image" / "pngb64str.txt", "r") as f:
        real_str = f.read().strip()
    assert png_str == real_str


def test_invalid_task_id():
    with pytest.raises(PartitionTaskNotFoundError):
        partition_file_async_result("INVALID_JOB_ID")


def test_partition_file_async_submit(mocker):
    data = b'{"task_id": "1234"}'
    expected_response = json.loads(data.decode())

    mocked_response = mocker.Mock()
    mocked_response.status_code = 202
    mocked_response.iter_bytes.return_value = (data,)

    mocker.patch("httpx.stream").return_value.__enter__.return_value = mocked_response

    with open(RESOURCE_DIR / "pdfs" / "3m_table.pdf", "rb") as f:
        response = partition_file_async_submit(f)

    assert response == expected_response


def test_partiton_file_async_url_forwarding(mocker):
    dummy = open("/dev/null", "rb")

    def call_partition_file(base_url: str):
        partition_file_async_submit(dummy, docparse_url=base_url)
        partition_file_async_submit(dummy, aps_url=base_url)
        partition_file_async_submit(
            dummy,
            aps_url="https://example.com/v1/document/partition",
            docparse_url=base_url,
        )
        partition_file_async_submit(dummy, aps_url=base_url, docparse_url=base_url)

    standard_async_url = ARYN_DOCPARSE_URL.replace("/v1/", "/v1/async/submit/")

    def check_standard_url(
        file: Union[BinaryIO, str, PathLike],
        **kwargs,
    ) -> None:
        url = kwargs.get("docparse_url") or kwargs.get("aps_url")
        assert url == standard_async_url

    mocker.patch(
        "aryn_sdk.client.partition._partition_file_inner",
        side_effect=check_standard_url,
    )
    partition_file_async_submit(dummy)
    call_partition_file(ARYN_DOCPARSE_URL)
    call_partition_file(standard_async_url)

    nonstandard_url_example = "http://localhost:8000/v1/document/partition"
    nonstandard_async_url_example = nonstandard_url_example.replace(
        "/v1/", "/v1/async/submit/"
    )

    def check_nonstandard_url(
        file: Union[BinaryIO, str, PathLike],
        **kwargs,
    ) -> None:
        url = kwargs.get("docparse_url") or kwargs.get("aps_url")
        assert url == nonstandard_async_url_example

    mocker.patch(
        "aryn_sdk.client.partition._partition_file_inner",
        side_effect=check_nonstandard_url,
    )
    call_partition_file(nonstandard_url_example)
    call_partition_file(nonstandard_async_url_example)

    dummy.close()


def test_partition_file_async_with_unsupported_file_format():
    with open(
        RESOURCE_DIR / "image" / "unsupported-format-test-document-image.heic", "rb"
    ) as f:
        task_id = partition_file_async_submit(f)["task_id"]

    start = time.time()
    while True:
        actual_result = partition_file_async_result(task_id)
        if (
            actual_result["task_status"] != "pending"
            or time.time() - start >= ASYNC_TIMEOUT
        ):
            break
        time.sleep(1)
    assert actual_result["task_status"] == "done"
    assert actual_result["result"] is not None
    assert actual_result["result"]["status_code"] == 500
    assert actual_result["result"]["error"] == "500: Failed to convert file to pdf"


def test_multiple_partition_file_async():
    num_tasks = 4
    task_ids = []

    before = partition_file_async_list()
    logging.info(f"List before:\n{json.dumps(before, indent=4)}")

    for i in range(num_tasks):
        logging.info(f"Submitting task {i + 1}/{num_tasks}")
        task_id = partition_file_async_submit(
            RESOURCE_DIR / "pdfs" / "FR-2002-05-03-TRUNCATED-40.pdf"
        )["task_id"]
        logging.info(f"\tTask ID: {task_id}")
        task_ids.append(task_id)

    after = partition_file_async_list()
    logging.info(f"List after:\n{json.dumps(after, indent=4)}")
    assert len(after) >= num_tasks
    for task_id in task_ids:
        assert task_id in after

    for i, task_id in enumerate(task_ids):
        logging.info(f"Polling task ({task_id}) {i + 1}/{num_tasks}")
        start = time.time()
        while True:
            actual_result = partition_file_async_result(task_id)
            if (
                actual_result["task_status"] != "pending"
                or time.time() - start >= ASYNC_TIMEOUT
            ):
                break
            time.sleep(1)
            logging.info(f"\tContinuing to Poll Task {task_id} ({i + 1}/{num_tasks})")
        assert actual_result["task_status"] == "done"
        assert len(actual_result["result"]["elements"]) > 1000


def test_partition_file_async_cancel():
    with open(RESOURCE_DIR / "pdfs" / "FR-2002-05-03-TRUNCATED-40.pdf", "rb") as f:
        task_id = partition_file_async_submit(f)["task_id"]

    before_cancel_result = partition_file_async_result(task_id)
    assert before_cancel_result["task_status"] == "pending"
    partition_file_async_cancel(task_id)

    # Cancellation is not reflected in the result immediately
    with pytest.raises(PartitionTaskNotFoundError):
        for _ in range(10):
            time.sleep(0.1)
            after_cancel_result = partition_file_async_result(task_id)
            if after_cancel_result["task_status"] != "pending":
                break
            assert after_cancel_result["task_status"] == "pending"


def test_smoke_webhook(mocker):
    data = b'{"task_id": "1234"}'

    webhook_url = "TEST"

    mocked_response = mocker.Mock()
    mocked_response.status_code = 202
    mocked_response.iter_bytes.return_value = data.split(sep=b"\n")

    def check_webhook(*args, request, **kwargs):
        assert "X-Aryn-Webhook" in request.headers
        assert request.headers.get("X-Aryn-Webhook") == webhook_url
        return mocked_response

    fake_post = mocker.patch("httpx.Client.send", side_effect=check_webhook)

    partition_file_async_submit(
        RESOURCE_DIR / "pdfs" / "3m_table.pdf", webhook_url=webhook_url
    )
    fake_post.assert_called()



from typing import Annotated, Literal, Optional, Union
from pydantic import BaseModel, Field, model_validator


class TokenizerOptions(BaseModel):
    model_name: Annotated[Optional[str], Field(description="The name of the tokenizer model.")] = None
    max_tokens: Annotated[Optional[int], Field(description="The maximum number of tokens.")] = None

    class Config:
        # Ensure that extra fields are not allowed
        extra = "forbid"


class ChunkingOptions(BaseModel):
    strategy: Annotated[
        Optional[Union[Literal["maximize_within_limit"], Literal["mixed_multi_column"], Literal["context_rich"]]],
        Field(
            description="The strategy to use for chunking. For convenience, when chunking options is specified but its strategy is unspecified, the service uses its default strategy."
        ),
    ] = None
    max_tokens: Annotated[Optional[int], Field(description="The maximum number of tokens per chunk.")] = None
    tokenizer: Annotated[
        Optional[Union[Literal["openai_tokenizer"], Literal["character_tokenizer"], Literal["huggingface_tokenizer"]]],
        Field(description="The tokenizer to use for chunking."),
    ] = None
    tokenizer_options: Annotated[Optional[TokenizerOptions], Field(description="Options for the tokenizer.")] = None
    merge_across_pages: Annotated[Optional[bool], Field(description="Whether to merge across pages.")] = None

    @model_validator(mode="after")
    def check_max_tokens_is_reasonable(self):
        if self.max_tokens is not None and self.max_tokens < 1:
            raise ValueError("When specified, max tokens must be at least 1.")
        return self

    @model_validator(mode="after")
    def check_merge_across_pages_applies(self):
        if self.merge_across_pages is not None and self.strategy == "mixed_multi_column":
            raise ValueError("Merge across pages option is not supported for mixed_multi_column strategy.")
        return self

    @model_validator(mode="after")
    def check_tokenizer_options_applies(self):
        if self.tokenizer == "character_tokenizer" and self.tokenizer_options is not None:
            raise ValueError("character_tokenizer takes no tokenizer_options.")

    class Config:
        # Ensure that extra fields are not allowed
        extra = "forbid"



from enum import StrEnum

from typing import Optional
from pydantic import BaseModel


class TaskStatus(StrEnum):
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    DONE = "DONE"
    CANCELLED = "CANCELLED"


class AsyncTaskMetadata(BaseModel):
    action: str
    task_status: TaskStatus
    userdata: Optional[str] = None


class AyncTaskList(BaseModel):
    tasks: list[AsyncTaskMetadata]



from datetime import datetime
from typing import Optional
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, JsonValue

from .prompt import PromptType
from .schema import Schema


class DocSetMetadata(BaseModel):
    model_config = ConfigDict(serialize_by_alias=True)

    account_id: str = Field(description="The account id containing the DocSet.")
    docset_id: str = Field(description="The unique id for the DocSet.")
    name: str = Field(description="The name of the DocSet.")
    created_at: datetime = Field(description="The creation time of this DocSet.")
    readonly: bool = Field(description="Whether the DocSet is read-only.")
    properties: Optional[dict[str, JsonValue]] = Field(
        default=None, description="Additional properties for the DocSet."
    )
    size: Optional[int] = Field(default=None, description="The size of the DocSet in bytes.")
    schema_: Optional[Schema] = Field(
        default=None,
        validation_alias=AliasChoices("query_schema", "schema"),
        serialization_alias="schema",
    )

    @property
    def query_schema(self) -> Optional[Schema]:
        return self.schema_

    # Map from prompt type to prompt_id.
    prompts: dict[PromptType, str] = Field(default={}, description="The prompts associated with this DocSet.")


class DocSetUpdate(BaseModel):
    model_config = ConfigDict(serialize_by_alias=True)

    name: Optional[str] = None
    properties: Optional[dict[str, JsonValue]] = None
    schema_: Optional[Schema] = Field(
        default=None,
        validation_alias=AliasChoices("query_schema", "schema"),
        serialization_alias="schema",
    )

    @property
    def query_schema(self) -> Optional[Schema]:
        return self.schema_

    prompts: Optional[dict[PromptType, str]] = None



import time

from typing import Generic, Optional, Type, TypeVar, TYPE_CHECKING

from .response import Response, SimpleResponse
from .exceptions import ArynSDKException

if TYPE_CHECKING:
    from aryn.client.client import Client

T = TypeVar("T")


class AsyncTask(Generic[T]):
    def __init__(
        self, client: "Client", task_id: str, method: Optional[str], path: Optional[str], response_type: Type[T]
    ) -> None:
        self.task_id = task_id
        self._method = method
        self._path = path
        self._client = client
        self._response_type = response_type

    @property
    def path(self) -> Optional[str]:
        return self._path

    @property
    def method(self) -> Optional[str]:
        return self._method

    def cancel(self) -> SimpleResponse:
        return self._client.cancel_async_task(self)

    # TODO: Figure out behavior if task is canceled
    def result(self, timeout: Optional[int] = None) -> Response[T]:
        """Returns the result of the task.

        Waits up to timeout seconds for the task to complete.
        """
        start_time = time.time()

        while True:
            res = self._client._get_async_result_internal(self)
            if res.status_code == 200:
                # Task has completed successfully.
                return Response(res, self._response_type(**res.json()))

            elif res.status_code == 202:
                # Task is still pending
                if timeout is not None and time.time() - start_time > timeout:
                    raise TimeoutError()
                time.sleep(1)

            else:
                # Task has failed. Should be unreachable.
                raise ArynSDKException(res)



from os import PathLike
import os
from pathlib import Path
from typing import Optional
import yaml

_DEFAULT_PATH = Path.home() / ".aryn" / "config.yaml"
_ARYN_API_KEY_ENV_VAR = "ARYN_API_KEY"
_ARYN_URL_ENV_VAR = "ARYN_API_URL"
_DEFAULT_ARYN_URL = "https://api.aryn.ai"


class ArynConfig:
    def __init__(
        self,
        aryn_config_path: PathLike = _DEFAULT_PATH,
        aryn_api_key: Optional[str] = None,
        aryn_url: Optional[str] = None,
    ):
        self._aryn_config_path = Path(aryn_config_path)
        self._aryn_api_key = aryn_api_key
        self._aryn_url = aryn_url

    def api_key(self) -> str:
        if self._aryn_api_key is not None:
            return self._aryn_api_key
        if val := os.environ.get(_ARYN_API_KEY_ENV_VAR):
            return val
        if Path(self._aryn_config_path).exists():
            with open(self._aryn_config_path, "r") as f:
                data = yaml.safe_load(f)
                if "aryn_token" in data:
                    return data["aryn_token"]
        raise ValueError(
            f"Could not find an aryn api key. Checked the {_ARYN_API_KEY_ENV_VAR} env "
            f"var, the {self._aryn_config_path} config file, and the aryn_api_key parameter"
        )

    def aryn_url(self) -> str:
        url = None

        if self._aryn_url is not None:
            url = self._aryn_url
        elif val := os.environ.get(_ARYN_URL_ENV_VAR):
            url = val
        elif Path(self._aryn_config_path).exists():
            with open(self._aryn_config_path, "r") as f:
                data = yaml.safe_load(f)
                if "aryn_url" in data:
                    url = data["aryn_url"]
        else:
            url = _DEFAULT_ARYN_URL

        if url is None:
            raise ValueError(
                f"Could not find an aryn url. Checked the {_ARYN_URL_ENV_VAR} env "
                f"var, the {self._aryn_config_path} config file, and the aryn_url parameter"
            )

        return url



from os import PathLike
from typing import BinaryIO, Union
import io
from packaging.version import InvalidVersion, Version
from pathlib import Path
import PIL
from PIL import Image, ImageDraw, ImageFont
import pdf2image
import logging
import sys


_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)
_logger.addHandler(logging.StreamHandler(sys.stderr))

COLORS = {
    "table": "tomato",
    "": "orange",  # table cells don't get a label bc it covers up the table label
    "Text": "blue",
    "Image": "green",
    "Section-header": "mediumblue",
    "Page-footer": "dimgrey",
    "List-item": "dodgerblue",
}


# Drawing algorithm, heavily adapted from sycamore.utils.pdf_utils.show_pages
# 1. Convert the pdf to a list of images
# 2. For each element
#     - find the page number it's on
#     - scale the bbox proportionally to the size of the page
#     - add a rectangle to the image of the page
#     - add a text box with the label to the image of the page
#     - if draw_table_cells and this element is a table
#         - do all the bbox/rectangle/textbox steps above for each cell.
def draw_with_boxes(
    pdf_file: Union[PathLike, BinaryIO, str],
    partitioning_data: dict,
    draw_table_cells: bool = False,
) -> list[Image.Image]:
    """
    Create a list of images from the provided pdf, one for each page, with bounding boxes detected by
    the partitioner drawn on.

    Args:
        pdf_file: an open file or path to a pdf file upon which to draw
        partitioning_data: the output from ``aryn_sdk.partition.partition_file``
        draw_table_cells: whether to draw individually detected cells of tables.
            default: False

    Returns:
        a list of images of pages of the pdf, each with bounding boxes drawn on

    Example:

         .. code-block:: python

            from aryn_sdk.partition import partition_file, draw_with_boxes

            with open("my-favorite-pdf.pdf", "rb") as f:
                data = partition_file(
                    f,
                    aryn_api_key="MY-ARYN-TOKEN",
                    use_ocr=True,
                    extract_table_structure=True,
                    extract_images=True
                )
            pages = draw_with_boxes("my-favorite-pdf.pdf", data, draw_table_cells=True)
    """
    assert (
        "elements" in partitioning_data
    ), 'There are no "elements" in the provided partitioning data, so drawing them is impossible'
    if isinstance(pdf_file, str):
        pdf_file = Path(pdf_file)
    if isinstance(pdf_file, PathLike):
        with open(pdf_file, "rb") as f:
            pdf_file = io.BytesIO(f.read())
    images = pdf2image.convert_from_bytes(pdf_file.read())
    for element in partitioning_data["elements"]:
        if "page_number" in element["properties"]:
            im = images[element["properties"]["page_number"] - 1]
            _draw_box_on_image(im, element)
            if draw_table_cells and element.get("type") == "table" and (table_object := element.get("table")):
                for cell in table_object.get("cells", []):
                    if cell and (bbox := cell.get("bbox")):
                        _draw_box_on_image(
                            im,
                            element={
                                "type": "",
                                "bbox": (
                                    bbox["x1"],
                                    bbox["y1"],
                                    bbox["x2"],
                                    bbox["y2"],
                                ),
                            },
                        )
    return images


def _supports_font_size() -> bool:
    try:
        return Version(PIL.__version__) >= Version("10.1.0")
    except InvalidVersion:
        return False


def _draw_box_on_image(image: Image.Image, element: dict):
    e_coords = element.get("bbox")
    if e_coords is None:
        return
    coords = (
        e_coords[0] * image.width,
        e_coords[1] * image.height,
        e_coords[2] * image.width,
        e_coords[3] * image.height,
    )
    label = element.get("type", "unknown")
    color = _color_for_label(label)
    canvas = ImageDraw.ImageDraw(image)
    canvas.rectangle(coords, outline=color, width=3)
    label_loc = (coords[0] - image.width / 100, coords[1] - image.height / 100)

    if _supports_font_size():
        font = ImageFont.load_default(size=20)
    else:
        font = ImageFont.load_default()

    font_box = canvas.textbbox(label_loc, label, font=font)
    canvas.rectangle(font_box, fill="yellow")
    canvas.text(label_loc, label, fill="black", font=font, align="left")


def _color_for_label(label) -> str:
    if label in COLORS:
        return COLORS[label]
    else:
        return "black"



from os import PathLike
from typing import BinaryIO, Literal, Optional, Union, Any
from urllib.parse import urlparse, urlunparse
from collections.abc import Mapping
from .config import ArynConfig
import httpx
import sys
import json
import logging
import pandas as pd
import numpy as np
from collections import OrderedDict
from PIL import Image
import base64
import io
import xml.etree.ElementTree as ET

# URL for Aryn DocParse
ARYN_DOCPARSE_URL = "https://api.aryn.cloud/v1/document/partition"

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)
_logger.addHandler(logging.StreamHandler(sys.stderr))

g_version = "0.2.5"
g_parameters = {"path_filter": "^/v1/document/partition$"}


class PartitionError(Exception):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code = status_code


class PartitionTaskError(Exception):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code = status_code


class PartitionTaskNotFoundError(Exception):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code = status_code


class BoolFlag:
    """
    A boxed boolean that can be mutated and passed around by reference.
    """

    __slots__ = "val"

    def __init__(self, val: bool) -> None:
        self.val = val

    def set(self, val: bool) -> None:
        self.val = val

    def get(self) -> bool:
        return self.val


def partition_file(
    file: Union[BinaryIO, str, PathLike],
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    threshold: Optional[Union[float, Literal["auto"]]] = None,
    text_mode: Optional[str] = None,
    use_ocr: Optional[bool] = None,
    summarize_images: bool = False,
    ocr_language: Optional[str] = None,
    extract_table_structure: bool = False,
    text_extraction_options: Optional[dict[str, Any]] = None,
    table_extraction_options: Optional[dict[str, Any]] = None,
    extract_images: bool = False,
    extract_image_format: Optional[str] = None,
    selected_pages: Optional[list[Union[list[int], int]]] = None,
    chunking_options: Optional[dict[str, Any]] = None,
    aps_url: Optional[str] = None,  # deprecated in favor of docparse_url
    docparse_url: Optional[str] = None,
    ssl_verify: bool = True,
    output_format: Optional[str] = None,
    markdown_options: Optional[dict[str, Any]] = None,
    output_label_options: Optional[dict[str, Any]] = None,
    trace_id: Optional[str] = None,  # deprecated
    extra_headers: Optional[dict[str, str]] = None,
    cancel_flag: Optional[BoolFlag] = None,
    add_to_docset_id: Optional[str] = None,
) -> dict:
    """
    Sends file to Aryn DocParse and returns a dict of its document structure and text

    Args:
        file: (pdf, docx, doc, jpg, or png, etc.) file to partition
            (see all supported file types at https://docs.aryn.ai/docparse/formats_supported)
        aryn_api_key: Aryn api key, provided as a string
            You can get a key here: https://www.aryn.ai/get-started
        aryn_config: ArynConfig object, used for finding an api key.
            If aryn_api_key is set it will override this.
            default: The default ArynConfig looks in the env var ARYN_API_KEY and the file ~/.aryn/config.yaml
        threshold: specify the cutoff for detecting bounding boxes. Must be set to "auto" or
            a floating point value between 0.0 and 1.0.
            default: None (Aryn DocParse will choose)
        text_mode:  Allows for specifying the text extraction mode. Valid options are 'standard', 'fine_grained', 'standard_ocr', and 'vision_ocr'.
            'standard' will use the standard text extraction pipeline, 'fine_grained' will use a more fine-grained text extraction pipeline,
            'standard_ocr' will use the standard OCR pipeline, and 'vision_ocr' will use a vision OCR pipeline. If this option is specified,
            it will override use_ocr and text_extraction_options even if specified.
            default: 'standard'
        use_ocr: deprecated, use text_mode instead.
            extract text using an OCR model instead of extracting embedded text in PDF.
            default: False
        summarize_images: Generate a text summary of detected images using a VLM.
        ocr_language: specify the language to use for OCR. If not set, the language will be english.
            default: English
        extract_table_structure: extract tables and their structural content.
            default: False
        text_extraction_options: 'ocr_text_mode' is deprecated, use text_mode instead.
            Specify options for text extraction, supports 'ocr_text_mode', with valid options 'vision' and 'standard' and boolean
            'remove_line_breaks'.For 'ocr_text_mode', attempt to extract all non-table text
            using vision models if 'vision', else will use the standard OCR pipeline. Vision is useful for documents with complex layouts
            or non-standard fonts. 'remove_line_breaks' will remove line breaks from the text.
            default: {'remove_line_breaks': False}
        table_extraction_options: Specify options for table extraction. Only enabled if table extraction
            is enabled. Default is {}. Options:
            - 'include_additional_text': Attempt to enhance the table structure by merging in tokens from
                text extraction. This can be useful for tables with missing or misaligned text. Default: False
            - 'model_selection': expression to instruct DocParse how to choose which model to use for table
                extraction. See https://docs.aryn.ai/docparse/processing_options for more details. Default:
                "pixels > 500 -> deformable_detr; table_transformer"
        extract_images: extract image contents. By default returns the image in base64 encoded ppm format,
              but can be configured to return the image in other formats via the extract_image_format parameter.
            default: False
        extract_image_format: specify the format of the extracted images. Only applies when extract_images=True.
              Must be one of ["PPM", "PNG", "JPEG"]
            default: PPM
        selected_pages: list of individual pages (1-indexed) from the pdf to partition
            default: None
        chunking_options: Specify options for chunking the document.
            You can use the default the chunking options by setting this to {}.
            Here is an example set of chunking options:
            {
                'strategy': 'context_rich',
                'tokenizer': 'openai_tokenizer',
                'tokenizer_options': {'model_name': 'text-embedding-3-small'},
                'max_tokens': 512,
                'merge_across_pages': True
            }
            default: None
        aps_url: url of Aryn DocParse endpoint.
            Left in for backwards compatibility. Use docparse_url instead.
        docparse_url: url of Aryn DocParse endpoint.
        ssl_verify: verify ssl certificates. In databricks, set this to False to fix ssl imcompatibilities.
        output_format: controls output representation; can be set to "markdown" or "json"
            default: None (JSON elements)
        markdown_options: A dictionary for configuring markdown output behavior. It supports three options:
            include_headers, a boolean specifying whether to include headers in the markdown output, include_footers,
            a boolean specifying whether to include footers in the markdown output, and include_pagenum, a boolean
            specifying whether to include page numbers in the markdown output. Here is an example set of markdown
            options:
                {
                    "include_headers": True,
                    "include_footers": True,
                    "include_pagenum": True
                }
        output_label_options: A dictionary for configuring output label behavior. It supports three options:
            promote_title, a boolean specifying whether to pick the largest element by font size on the first page
                from among the elements on that page that have one of the types specified in title_candidate_elements
                and promote it to type "Title" if there is no element on the first page of type "Title" already.
            title_candidate_elements, a list of strings representing the label types allowed to be promoted to
                a title.
            orientation_correction, a boolean specifying whether to pagewise rotate pages to the correct orientation
                based off the orientation of text. Pages are rotated by increments of 90 degrees to correct their
                orientation.
            Here is an example set of output label options:
                {
                    "promote_title": True,
                    "title_candidate_elements": ["Section-header", "Caption"],
                    "orientation_correction": True
                }
            default: None (no element is promoted to "Title")
        trace_id: deprecated
        extra_headers: dict of HTTP headers to send to DocParse
        cancel_flag: way to interrupt partitioning from the outside
        add_to_docset_id: An optional docset_id to which to add the document.

    Returns:
        A dictionary containing "status", "elements", and possibly "error"
        If output_format is "markdown" then it returns a dictionary of "status", "markdown", and possibly "error"

    Example:
         .. code-block:: python

            from aryn_sdk.partition import partition_file

            with open("my-favorite-pdf.pdf", "rb") as f:
                data = partition_file(
                    f,
                    aryn_api_key="MY-ARYN-API-KEY",
                    text_mode="standard_ocr",
                    extract_table_structure=True,
                    extract_images=True
                )
            elements = data['elements']
    """

    return _partition_file_wrapper(
        file=file,
        aryn_api_key=aryn_api_key,
        aryn_config=aryn_config,
        threshold=threshold,
        text_mode=text_mode,
        use_ocr=use_ocr,
        summarize_images=summarize_images,
        ocr_language=ocr_language,
        extract_table_structure=extract_table_structure,
        text_extraction_options=text_extraction_options,
        table_extraction_options=table_extraction_options,
        extract_images=extract_images,
        extract_image_format=extract_image_format,
        selected_pages=selected_pages,
        chunking_options=chunking_options,
        aps_url=aps_url,
        docparse_url=docparse_url,
        ssl_verify=ssl_verify,
        output_format=output_format,
        markdown_options=markdown_options,
        output_label_options=output_label_options,
        trace_id=trace_id,
        extra_headers=extra_headers,
        cancel_flag=cancel_flag,
        add_to_docset_id=add_to_docset_id,
    )


def _partition_file_wrapper(
    file: Union[BinaryIO, str, PathLike],
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    threshold: Optional[Union[float, Literal["auto"]]] = None,
    text_mode: Optional[str] = None,
    use_ocr: Optional[bool] = None,
    summarize_images: bool = False,
    ocr_language: Optional[str] = None,
    extract_table_structure: bool = False,
    text_extraction_options: Optional[dict[str, Any]] = None,
    table_extraction_options: Optional[dict[str, Any]] = None,
    extract_images: bool = False,
    extract_image_format: Optional[str] = None,
    selected_pages: Optional[list[Union[list[int], int]]] = None,
    chunking_options: Optional[dict[str, Any]] = None,
    aps_url: Optional[str] = None,  # deprecated in favor of docparse_url
    docparse_url: Optional[str] = None,
    ssl_verify: bool = True,
    output_format: Optional[str] = None,
    markdown_options: Optional[dict[str, Any]] = None,
    output_label_options: Optional[dict[str, Any]] = None,
    webhook_url: Optional[str] = None,
    trace_id: Optional[str] = None,  # deprecated
    extra_headers: Optional[dict[str, str]] = None,
    cancel_flag: Optional[BoolFlag] = None,
    add_to_docset_id: Optional[str] = None,
):
    """Do not call this function directly. Use partition_file or partition_file_async_submit instead."""

    # If you hand me a path for the file, read it in instead of trying to send the path
    should_close = False
    try:
        if isinstance(file, (str, PathLike)):
            file = open(file, "rb")
            should_close = True
        return _partition_file_inner(
            file=file,
            aryn_api_key=aryn_api_key,
            aryn_config=aryn_config,
            threshold=threshold,
            text_mode=text_mode,
            use_ocr=use_ocr,
            summarize_images=summarize_images,
            ocr_language=ocr_language,
            extract_table_structure=extract_table_structure,
            text_extraction_options=text_extraction_options,
            table_extraction_options=table_extraction_options,
            extract_images=extract_images,
            extract_image_format=extract_image_format,
            selected_pages=selected_pages,
            chunking_options=chunking_options,
            aps_url=aps_url,
            docparse_url=docparse_url,
            ssl_verify=ssl_verify,
            output_format=output_format,
            markdown_options=markdown_options,
            output_label_options=output_label_options,
            trace_id=trace_id,
            extra_headers=extra_headers,
            cancel_flag=cancel_flag,
            webhook_url=webhook_url,
            add_to_docset_id=add_to_docset_id,
        )
    finally:
        if should_close and isinstance(file, BinaryIO):
            file.close()


def _partition_file_inner(
    file: BinaryIO,
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    threshold: Optional[Union[float, Literal["auto"]]] = None,
    text_mode: Optional[str] = None,
    use_ocr: Optional[bool] = None,
    summarize_images: bool = False,
    ocr_language: Optional[str] = None,
    extract_table_structure: bool = False,
    text_extraction_options: Optional[dict[str, Any]] = None,
    table_extraction_options: Optional[dict[str, Any]] = None,
    extract_images: bool = False,
    extract_image_format: Optional[str] = None,
    selected_pages: Optional[list[Union[list[int], int]]] = None,
    chunking_options: Optional[dict[str, Any]] = None,
    aps_url: Optional[str] = None,  # deprecated in favor of docparse_url
    docparse_url: Optional[str] = None,
    ssl_verify: bool = True,
    output_format: Optional[str] = None,
    markdown_options: Optional[dict[str, Any]] = None,
    output_label_options: Optional[dict[str, Any]] = None,
    trace_id: Optional[str] = None,  # deprecated
    extra_headers: Optional[dict[str, str]] = None,
    cancel_flag: Optional[BoolFlag] = None,
    webhook_url: Optional[str] = None,
    add_to_docset_id: Optional[str] = None,
):
    """Do not call this function directly. Use partition_file or partition_file_async_submit instead."""

    aryn_config = _process_config(aryn_api_key, aryn_config)

    if aps_url is not None:
        if docparse_url is not None:
            logging.warning(
                '"aps_url" and "docparse_url" parameters were both set. "aps_url" is deprecated. Using "docparse_url".'
            )
        else:
            logging.warning('"aps_url" parameter is deprecated. Use "docparse_url" instead')
            docparse_url = aps_url
    if text_mode is not None and (
        use_ocr is not None or (text_extraction_options and ("ocr_text_mode" in text_extraction_options))
    ):
        logging.warning(
            '"text_mode" parameter was set. Since "use_ocr" and "ocr_text_mode" parameters are deprecated, using "text_mode".'
        )
    else:
        if use_ocr:
            ocr_text_mode = "standard"
            if text_extraction_options and ("ocr_text_mode" in text_extraction_options):
                ocr_text_mode = text_extraction_options["ocr_text_mode"]
            text_mode = f"{ocr_text_mode}_ocr"
    if docparse_url is None:
        docparse_url = ARYN_DOCPARSE_URL
    source = "aryn-sdk"
    if extra_headers:
        source = extra_headers.get("X-Aryn-Origin", "aryn-sdk")

    options_str = _json_options(
        threshold=threshold,
        text_mode=text_mode,
        summarize_images=summarize_images,
        ocr_language=ocr_language,
        extract_table_structure=extract_table_structure,
        text_extraction_options=text_extraction_options,
        table_extraction_options=table_extraction_options,
        extract_images=extract_images,
        extract_image_format=extract_image_format,
        selected_pages=selected_pages,
        output_format=output_format,
        chunking_options=chunking_options,
        markdown_options=markdown_options,
        output_label_options=output_label_options,
        add_to_docset_id=add_to_docset_id,
        source=source,
    )

    _logger.debug(f"{options_str}")

    files: Mapping = {"options": options_str.encode("utf-8"), "pdf": file}
    headers = _generate_headers(aryn_config.api_key(), webhook_url, trace_id, extra_headers)

    content = []
    partial_line = []
    in_bulk = False

    # If you run into issues VCR.py, you can apply the monkey patch described here:
    # https://github.com/kevin1024/vcrpy/issues/656#issuecomment-2492379346
    with httpx.stream("POST", docparse_url, files=files, headers=headers, verify=ssl_verify, timeout=500) as resp:
        raise_error_on_non_2xx(resp)

        for part in resp.iter_bytes():
            # A big doc could take a while; we may be asked to bail out early
            if cancel_flag and cancel_flag.get():
                resp.close()
                break

            if not part:
                continue

            content.append(part)
            if in_bulk:
                continue

            partial_line.append(part)
            if b"\n" not in part:
                continue

            these_lines = b"".join(partial_line).split(b"\n")
            partial_line = [these_lines.pop()]

            for line in these_lines:
                if line.startswith(b"  ],"):
                    in_bulk = True
                    break
                if line.startswith(b'    "T+'):
                    t = json.loads(line.decode("utf-8").removesuffix(","))
                    _logger.info(f"ArynPartitioner: {t}")

    body = b"".join(content).decode("utf-8")
    _logger.debug("Recieved data from ArynPartitioner")

    data = json.loads(body)
    assert isinstance(data, dict)
    status = data.get("status", [])
    if error := data.get("error"):
        code = data.get("status_code")
        if code is None:
            code = 429 if error.startswith("429: ") else 500
        if code == 429:
            prefix = "Limit exceeded"
        else:
            prefix = "Error partway through processing"
        _logger.info(f"Error from ArynPartitioner: {error}")
        raise PartitionError(f"{prefix}: {error}\nPartial Status:\n{status}", code)
    return data


def raise_error_on_non_2xx(resp: httpx.Response) -> None:
    if resp.status_code < 200 or resp.status_code > 299:
        resp.read()  # Try to read the reponse to get the error message
        raise httpx.HTTPStatusError(
            f"Error: status_code: {resp.status_code}, reason: {resp.text}",
            request=resp.request,
            response=resp,
        )


def _process_config(aryn_api_key: Optional[str] = None, aryn_config: Optional[ArynConfig] = None) -> ArynConfig:
    if aryn_api_key is not None:
        if aryn_config is not None:
            _logger.warning("Both aryn_api_key and aryn_config were provided. Using aryn_api_key")
        aryn_config = ArynConfig(aryn_api_key=aryn_api_key)
    if aryn_config is None:
        aryn_config = ArynConfig()
    return aryn_config


def _generate_headers(
    aryn_api_key: str,
    webhook_url: Optional[str] = None,
    trace_id: Optional[str] = None,  # deprecated
    extra_headers: Optional[dict[str, str]] = None,
) -> dict[str, str]:
    # Start with user-supplied headers so they can't stomp official ones.
    headers = extra_headers.copy() if extra_headers else {}
    if webhook_url:
        headers["X-Aryn-Webhook"] = webhook_url
    if trace_id:
        headers["X-Aryn-Trace-ID"] = trace_id  # deprecated
    headers["Authorization"] = f"Bearer {aryn_api_key}"
    headers["User-Agent"] = f"aryn-sdk/{g_version}"
    return headers


def _json_options(
    threshold: Optional[Union[float, Literal["auto"]]] = None,
    text_mode: Optional[str] = None,
    summarize_images: bool = False,
    ocr_language: Optional[str] = None,
    extract_table_structure: bool = False,
    text_extraction_options: Optional[dict[str, Any]] = None,
    table_extraction_options: Optional[dict[str, Any]] = None,
    extract_images: bool = False,
    extract_image_format: Optional[str] = None,
    selected_pages: Optional[list[Union[list[int], int]]] = None,
    output_format: Optional[str] = None,
    chunking_options: Optional[dict[str, Any]] = None,
    markdown_options: Optional[dict[str, Any]] = None,
    output_label_options: Optional[dict[str, Any]] = None,
    add_to_docset_id: Optional[str] = None,
    source: str = "aryn-sdk",
) -> str:
    # isn't type-checking fun
    options: dict[str, Union[float, bool, str, list[Union[list[int], int]], dict[str, Any]]] = dict()
    if threshold is not None:
        options["threshold"] = threshold
    if summarize_images:
        options["summarize_images"] = summarize_images
    if ocr_language:
        options["ocr_language"] = ocr_language
    if extract_images:
        options["extract_images"] = extract_images
    if extract_image_format:
        options["extract_image_format"] = extract_image_format
    if text_mode:
        options["text_mode"] = text_mode
    if extract_table_structure:
        options["extract_table_structure"] = extract_table_structure
    if text_extraction_options:
        options["text_extraction_options"] = text_extraction_options
    if table_extraction_options:
        options["table_extraction_options"] = table_extraction_options
    if selected_pages:
        options["selected_pages"] = selected_pages
    if output_format:
        options["output_format"] = output_format
    if chunking_options is not None:
        options["chunking_options"] = chunking_options
    if markdown_options:
        options["markdown_options"] = markdown_options
    if output_label_options:
        options["output_label_options"] = output_label_options
    if add_to_docset_id:
        options["add_to_docset_id"] = add_to_docset_id

    options["source"] = source

    return json.dumps(options)


def partition_file_async_submit(
    file: Union[BinaryIO, str, PathLike],
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    threshold: Optional[Union[float, Literal["auto"]]] = None,
    text_mode: Optional[str] = None,
    use_ocr: Optional[bool] = None,
    summarize_images: bool = False,
    ocr_language: Optional[str] = None,
    extract_table_structure: bool = False,
    text_extraction_options: Optional[dict[str, Any]] = None,
    table_extraction_options: Optional[dict[str, Any]] = None,
    extract_images: bool = False,
    extract_image_format: Optional[str] = None,
    selected_pages: Optional[list[Union[list[int], int]]] = None,
    chunking_options: Optional[dict[str, Any]] = None,
    aps_url: Optional[str] = None,  # deprecated in favor of docparse_url
    docparse_url: Optional[str] = None,
    ssl_verify: bool = True,
    output_format: Optional[str] = None,
    markdown_options: Optional[dict[str, Any]] = None,
    output_label_options: Optional[dict[str, Any]] = None,
    trace_id: Optional[str] = None,  # deprecated
    extra_headers: Optional[dict[str, str]] = None,
    webhook_url: Optional[str] = None,
    async_submit_url: Optional[str] = None,
    add_to_docset_id: Optional[str] = None,
) -> dict[str, Any]:
    """
    Submits a file to be partitioned asynchronously. Meant to be used in tandem with `partition_file_async_result`.

    `partition_file_async_submit` takes the same arguments as `partition_file`, and in addition it accepts a str
    `webhook_url` argument which is a URL Aryn will send a POST request to when the task stops and an str
    `async_submit_url` argument that can be used to override where the task is submitted to.

    Set the `docparse_url` argument to the url of the synchronous endpoint, and this function will automatically
    change it to the async endpoint as long as `async_submit_url` is not set.

    For examples of usage see README.md

    Args:
        Includes All Arguments `partition_file` accepts plus those below:
        ...
        webhook_url: A URL to send a POST request to when the task is done. The resulting POST request will have a
            body like: {"done": [{"task_id": "aryn:j-47gpd3604e5tz79z1jro5fc"}]}
        async_submit_url: When set, this will override the endpoint the task is submitted to.

    Returns:
        A dictionary containing the key "task_id" the value of which can be used with the `partition_file_async_result`
        function to get the results and check the status of the async task.
    """

    if async_submit_url:
        docparse_url = async_submit_url
    elif not aps_url and not docparse_url:
        docparse_url = _convert_sync_to_async_url(ARYN_DOCPARSE_URL, "/submit", truncate=False)
    else:
        if aps_url:
            aps_url = _convert_sync_to_async_url(aps_url, "/submit", truncate=False)
        if docparse_url:
            docparse_url = _convert_sync_to_async_url(docparse_url, "/submit", truncate=False)

    return _partition_file_wrapper(
        file=file,
        aryn_api_key=aryn_api_key,
        aryn_config=aryn_config,
        threshold=threshold,
        text_mode=text_mode,
        use_ocr=use_ocr,
        summarize_images=summarize_images,
        ocr_language=ocr_language,
        extract_table_structure=extract_table_structure,
        text_extraction_options=text_extraction_options,
        table_extraction_options=table_extraction_options,
        extract_images=extract_images,
        extract_image_format=extract_image_format,
        selected_pages=selected_pages,
        chunking_options=chunking_options,
        aps_url=aps_url,
        docparse_url=docparse_url,
        ssl_verify=ssl_verify,
        output_format=output_format,
        markdown_options=markdown_options,
        output_label_options=output_label_options,
        trace_id=trace_id,
        extra_headers=extra_headers,
        webhook_url=webhook_url,
        add_to_docset_id=add_to_docset_id,
    )


def _convert_sync_to_async_url(url: str, prefix: str, *, truncate: bool) -> str:
    parsed_url = urlparse(url)
    assert parsed_url.path.startswith("/v1/")
    if parsed_url.path.startswith("/v1/async/submit"):
        return url
    ary = list(parsed_url)
    if truncate:
        ary[2] = f"/v1/async{prefix}"  # path
    else:
        ary[2] = f"/v1/async{prefix}{parsed_url.path[3:]}"  # path
    return urlunparse(ary)


def partition_file_async_result(
    task_id: str,
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    ssl_verify: bool = True,
    async_result_url: Optional[str] = None,
) -> dict[str, Any]:
    """
    Get the results of an asynchronous partitioning task by task_id. Meant to be used with
    `partition_file_async_submit`.

    For examples of usage see README.md

    Raises a `PartitionTaskNotFoundError` if the not task with the task_id can be found.

    Returns:
        A dict containing "status" and "status_code". When "status" is "done", the returned dict also contains "result"
        which contains what would have been returned had `partition_file` been called directly. "status" can be "done"
        or "pending".

        Unlike `partition_file`, this function does not raise an Exception if the partitioning failed.
    """
    if not async_result_url:
        async_result_url = _convert_sync_to_async_url(ARYN_DOCPARSE_URL, "/result", truncate=True)

    aryn_config = _process_config(aryn_api_key, aryn_config)

    specific_task_url = f"{async_result_url.rstrip('/')}/{task_id}"
    headers = _generate_headers(aryn_config.api_key())
    response = httpx.get(specific_task_url, params=g_parameters, headers=headers, verify=ssl_verify)
    if response.status_code == 200:
        return {"task_status": "done", "result": response.json()}
    elif response.status_code == 202:
        return {"task_status": "pending"}
    elif response.status_code == 404:
        raise PartitionTaskNotFoundError("No such task", response.status_code)
    else:
        raise_error_on_non_2xx(response)
        raise PartitionTaskError("Unexpected response code", response.status_code)


def partition_file_async_cancel(
    task_id: str,
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    ssl_verify: bool = True,
    async_cancel_url: Optional[str] = None,
) -> None:
    """
    Cancel an asynchronous partitioning task by task_id. Meant to be used with `partition_file_async_submit`.

    Raises an exception if there is no cancellable task found with the given task_id. A task can only be successfully
    cancelled once.

    For an example of usage see README.md
    """
    if not async_cancel_url:
        async_cancel_url = _convert_sync_to_async_url(ARYN_DOCPARSE_URL, "/cancel", truncate=True)

    aryn_config = _process_config(aryn_api_key, aryn_config)

    specific_task_url = f"{async_cancel_url.rstrip('/')}/{task_id}"
    headers = _generate_headers(aryn_config.api_key())
    response = httpx.post(specific_task_url, params=g_parameters, headers=headers, verify=ssl_verify)
    if response.status_code == 200:
        return
    elif response.status_code == 404:
        raise PartitionTaskNotFoundError("No such task", response.status_code)
    else:
        raise_error_on_non_2xx(response)
        raise PartitionTaskError("Unexpected response code.", response.status_code)


def partition_file_async_list(
    *,
    aryn_api_key: Optional[str] = None,
    aryn_config: Optional[ArynConfig] = None,
    ssl_verify: bool = True,
    async_list_url: Optional[str] = None,
) -> dict[str, Any]:
    """
    List pending async tasks. For an example of usage see README.md

    Returns:
        A dict like the one below which maps task_ids to a dict containing details of the respective task.

        {
            "aryn:j-sc0v0lglkauo774pioflp4l": {
                "state": "run"
            },
            "aryn:j-b9xp7ny0eejvqvbazjhg8rn": {
                "state": "run"
            }
        }
    """
    if not async_list_url:
        async_list_url = _convert_sync_to_async_url(ARYN_DOCPARSE_URL, "/list", truncate=True)

    aryn_config = _process_config(aryn_api_key, aryn_config)

    headers = _generate_headers(aryn_config.api_key())
    response = httpx.get(async_list_url, params=g_parameters, headers=headers, verify=ssl_verify)
    raise_error_on_non_2xx(response)
    if response.status_code != 200:
        raise PartitionTaskError("Unexpected response code", response.status_code)

    result = response.json()

    tasks = result["tasks"]
    for v in tasks.values():
        v.pop("action", None)
    return tasks


def table_elem_to_html(elem: dict[str, Any], pretty: bool = False) -> Optional[str]:
    """
    Create an HTML table representing the tabular data inside the provided table element.
    If the element is not of type 'table' or doesn't contain any table data, return None instead.

    Args:
        elem: An element dict from the 'elements' field of a ``partition_file`` response.
        pretty: If True, pretty print the HTML output. Defaults to false.

    Example:
         .. code-block:: python

            from aryn_sdk.partition import partition_file, table_elem_to_html

            with open("partition-me.pdf", "rb") as f:
                data = partition_file(
                    f,
                    text_mode="standard_ocr",
                    extract_table_structure=True,
                    extract_images=True
                )

            # Find the first table and convert it to HTML
            html = None
            for element in data['elements']:
                if element['type'] == 'table':
                    html = table_elem_to_html(element)
                    break

    """
    if elem["type"] != "table" or elem["table"] is None:
        return None

    table = ET.Element("table")
    root = table

    table_elem = elem["table"]

    curr_row = -1
    row = None

    if table_elem["caption"] is not None:
        caption = ET.SubElement(table, "caption")
        caption.text = table_elem["caption"]

    for cell in table_elem["cells"]:
        cell_attribs = {}

        rowspan = len(cell["rows"])
        colspan = len(cell["cols"])

        if rowspan > 1:
            cell_attribs["rowspan"] = str(rowspan)
        if colspan > 1:
            cell_attribs["colspan"] = str(colspan)

        if cell["rows"][0] > curr_row:
            curr_row = cell["rows"][0]
            row = ET.SubElement(table, "tr")

        assert row is not None
        tcell = ET.SubElement(row, "th" if cell["is_header"] else "td", attrib=cell_attribs)
        tcell.text = cell["content"]

    if pretty:
        ET.indent(root)

    return ET.tostring(root, encoding="unicode")


# Heavily adapted from lib/sycamore/data/table.py::Table.to_csv()
def table_elem_to_dataframe(elem: dict) -> Optional[pd.DataFrame]:
    """
    Create a pandas DataFrame representing the tabular data inside the provided table element.
    If the element is not of type 'table' or doesn't contain any table data, return None instead.

    Args:
        elem: An element from the 'elements' field of a ``partition_file`` response.

    Example:
         .. code-block:: python

            from aryn_sdk.partition import partition_file, table_elem_to_dataframe

            with open("partition-me.pdf", "rb") as f:
                data = partition_file(
                    f,
                    text_mode="standard_ocr",
                    extract_table_structure=True,
                    extract_images=True
                )

            # Find the first table and convert it to a dataframe
            df = None
            for element in data['elements']:
                if element['type'] == 'table':
                    df = table_elem_to_dataframe(element)
                    break
    """

    if (elem["type"] != "table") or (elem["table"] is None):
        return None

    table = elem["table"]

    header_rows = sorted(set(row_num for cell in table["cells"] for row_num in cell["rows"] if cell["is_header"]))
    i = -1
    for i in range(len(header_rows)):
        if header_rows[i] != i:
            break
    max_header_prefix_row = i
    grid_width = table["num_cols"]
    grid_height = table["num_rows"]

    grid = np.empty([grid_height, grid_width], dtype="object")
    for cell in table["cells"]:
        if cell["is_header"] and cell["rows"][0] <= max_header_prefix_row:
            for col in cell["cols"]:
                grid[cell["rows"][0], col] = cell["content"]
            for row in cell["rows"][1:]:
                for col in cell["cols"]:
                    grid[row, col] = ""
        else:
            grid[cell["rows"][0], cell["cols"][0]] = cell["content"]
            for col in cell["cols"][1:]:
                grid[cell["rows"][0], col] = ""
            for row in cell["rows"][1:]:
                for col in cell["cols"]:
                    grid[row, col] = ""

    header = grid[: max_header_prefix_row + 1, :]
    flattened_header = []
    for npcol in header.transpose():
        flattened_header.append(" | ".join(OrderedDict.fromkeys((c for c in npcol if c != ""))))
    df = pd.DataFrame(
        grid[max_header_prefix_row + 1 :, :],
        index=None,
        columns=flattened_header if max_header_prefix_row >= 0 else None,
    )

    return df


def tables_to_pandas(data: dict) -> list[tuple[dict, Optional[pd.DataFrame]]]:
    """
    For every table element in the provided partitioning response, create a pandas
    DataFrame representing the tabular data. Return a list containing all the elements,
    with tables paired with their corresponding DataFrames.

    Args:
        data: a response from ``partition_file``

    Example:
         .. code-block:: python

            from aryn_sdk.partition import partition_file, tables_to_pandas

            with open("my-favorite-pdf.pdf", "rb") as f:
                data = partition_file(
                    f,
                    aryn_api_key="MY-ARYN-API-KEY",
                    text_mode="standard_ocr",
                    extract_table_structure=True,
                    extract_images=True
                )
            elts_and_dataframes = tables_to_pandas(data)

    """
    results = []
    for e in data["elements"]:
        results.append((e, table_elem_to_dataframe(e)))

    return results


def tables_to_html(data: dict) -> list[tuple[dict, Optional[str]]]:
    """
    For every table element in the provided partitioning response, create an HTML
    string representing the tabular data. Return a list containing all the elements,
    with tables paired with their corresponding HTML.

    Args:
        data: a response from ``partition_file``

    Example:
         .. code-block:: python

            from aryn_sdk.partition import partition_file, tables_to_html

            with open("my-favorite-pdf.pdf", "rb") as f:
                data = partition_file(
                    f,
                    aryn_api_key="MY-ARYN-API-KEY",
                    text_mode="standard_ocr",
                    extract_table_structure=True,
                    extract_images=True
                )
            elts_and_html = tables_to_html(data)
    """
    results = []
    for e in data["elements"]:
        results.append((e, table_elem_to_html(e)))

    return results


def convert_image_element(
    elem: dict, format: str = "PIL", b64encode: bool = False
) -> Optional[Union[Image.Image, bytes, str]]:
    """
    Convert an image element to a more useable format. If no format is specified,
    create a PIL Image object. If a format is specified, output the bytes of the image
    in that format. If b64encode is set to True, base64-encode the bytes and return them
    as a string.

    Args:
        elem: an image element from the 'elements' field of a ``partition_file`` response
        format: an optional format to output bytes of. Default is PIL
        b64encode: base64-encode the output bytes. Format must be set to use this

    Example:
         .. code-block:: python

            from aryn_sdk.partition import partition_file, convert_image_element

            with open("my-favorite-pdf.pdf", "rb") as f:
                data = partition_file(
                    f,
                    extract_images=True
                )
            image_elts = [e for e in data['elements'] if e['type'] == 'Image']

            pil_img = convert_image_element(image_elts[0])
            jpg_byes = convert_image_element(image_elts[1], format='JPEG')
            png_str = convert_image_element(image_elts[2], format="PNG", b64encode=True)

    """
    if b64encode and format == "PIL":
        raise ValueError("b64encode was True but format was PIL. Cannot b64-encode a PIL Image")

    if elem.get("type") != "Image":
        return None

    width = elem["properties"]["image_size"][0]
    height = elem["properties"]["image_size"][1]
    mode = elem["properties"]["image_mode"]

    raw_bytes = base64.b64decode(elem["binary_representation"])
    in_format = elem["properties"].get("image_format")

    if in_format is None:
        im = Image.frombytes(mode, (width, height), data=raw_bytes)
    else:
        in_buf = io.BytesIO(raw_bytes)
        im = Image.open(in_buf, formats=[in_format])

    if format == "PIL":
        return im

    buf = io.BytesIO()
    im.save(buf, format)

    if not b64encode:
        return buf.getvalue()
    else:
        return base64.b64encode(buf.getvalue()).decode("utf-8")



from enum import Enum
from typing import Any, MutableMapping, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny, model_validator
from pydantic.json_schema import SkipJsonSchema

from .document import Document


class Node(BaseModel):
    """A single operation in a logical plan."""

    model_config = ConfigDict(extra="allow", use_attribute_docstrings=True)

    node_type: Optional[str] = Field(default=None)
    """The type of this node."""

    node_id: int
    """A unique integer ID representing this node."""

    description: Optional[str] = Field(None, json_schema_extra={"exclude_from_comparison": True})
    """A detailed description of why this operator was chosen for this query plan."""

    inputs: list[int] = []
    """A list of node IDs that this operation depends on."""


class LogicalPlan(BaseModel):
    """A logical query plan used to evaluate a query."""

    model_config = ConfigDict(use_attribute_docstrings=True)

    query: str
    """The query that the plan is for."""

    nodes: MutableMapping[int, SerializeAsAny[Node]]
    """A mapping of node IDs to nodes in the query plan."""

    result_node: int
    """The ID of the node that is the result of the query."""

    llm_prompt: Optional[Any] = None
    """The LLM prompt that was used to generate this query plan."""

    llm_plan: Optional[str] = None
    """The result generated by the LLM."""


class Query(BaseModel):
    """A query against a DocSet. Contains either a natural language string or a query plan."""

    model_config = ConfigDict(use_attribute_docstrings=True)

    docset_id: Optional[str] = None
    """The docset against which to run the query"""

    query: Optional[str] = None
    """The natural language query to run. if specified, `plan` must not be set."""

    plan: Optional[LogicalPlan] = None
    """The logical query plan to run. If specified, `query` must not be set."""

    stream: bool = False
    """If true, query results will be streamed back to the client as they are generated. Applies only when calling the query api."""

    summarize_result: bool = False
    """
    If true, an english summary of the result in context of the original query will be returned.
    Applies only when calling the query API and only available when `stream=True`
    """

    # Bookmarks are currently tied to the UI, so we hide them from the
    # generated OpenAPI schema so as not to confuse callers of the API.
    bookmark_source: SkipJsonSchema[Optional[str]] = None

    bookmark_target: SkipJsonSchema[Optional[str]] = None

    @model_validator(mode="after")
    def check_not_both_query_and_plan(self):
        if self.query is not None and self.plan is not None:
            raise ValueError("query and plan cannot both be specified")
        if self.query is None and self.plan is None:
            raise ValueError("one of query or plan is required")
        return self


class QueryResult(BaseModel):
    """The result of a non-streaming query."""

    model_config = ConfigDict(use_attribute_docstrings=True)

    query_id: str
    """The unique ID of the query operation."""

    plan: LogicalPlan
    """The logical query plan that was executed."""

    result: Any
    """The result of the query operation. Depending on the query, this could be a list of documents,
    a single document, a string, an integer, etc.
    """


class QueryTraceDoc(BaseModel):
    """A document in the trace of a query result."""

    model_config = ConfigDict(use_attribute_docstrings=True)

    node_id: int
    """The ID of the node in the query plan that produced this document."""

    doc: dict[str, Any]
    """The document data."""


class QueryEventType(str, Enum):
    """The type of event that occurred in the query trace."""

    COMPLETE = "complete"
    ERROR = "error"
    PLAN = "plan"
    RESULT = "result"
    RESULT_DOC = "result_doc"
    STATUS = "status"
    TRACE_DOC = "trace_doc"


class QueryEvent(BaseModel):
    """An event in the trace of a query result."""

    event_type: QueryEventType
    """The type of event."""

    data: Union[int, str, float, QueryTraceDoc, LogicalPlan, QueryResult, Document]
    """The data associated with the event."""



class ArynSDKException(Exception):
    def __init__(self, raw_response):
        self.raw_response = raw_response

    @property
    def status_code(self):
        return self.raw_response.status_code

    def __str__(self):
        return f"ArynSDKException: status_code: {self.status_code}, raw_response: {self.raw_response}"



from enum import StrEnum

from typing import Annotated, Optional
from pydantic import BaseModel, Field


class TaskStatus(StrEnum):
    QUEUED = "queued"
    RUNNING = "running"
    DONE = "done"
    CANCELLED = "cancelled"

    @classmethod
    def _missing_(cls, value):
        value = value.lower()  # type: ignore
        for member in cls:
            if member.value == value:
                return member
        return None


class AsyncTaskMetadata(BaseModel):
    """Metadata associated with an async task."""

    action: Annotated[str, Field(description="The path of the endpoint used to invoke the task.")]
    task_status: Annotated[TaskStatus, Field(description="The current status of the task.")]
    userdata: Annotated[Optional[str], Field(description="Additional metadata associated with the task.")] = None


class AsyncTaskMap(BaseModel):
    """Metadata describing a list of async tasks."""

    tasks: Annotated[dict[str, AsyncTaskMetadata], Field(description="A map of task ids to metadata for async tasks.")]
